/* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */

#include "rb-papi.h"

static ID id_new, id_begin, id_end, id_xres, id_yres;
static ID id_units, id_tv_sec, id_value;
static VALUE cAttribute, cResolution, cMetadata;

static void
no_free(void *ptr)
{
}

VALUE
rb_papi_c2r_attribute_no_free(papi_attribute_t *attr)
{
    return Data_Wrap_Struct(cAttribute, 0, no_free, attr);
}

static VALUE
rb_papi_attribute_get_name(VALUE self)
{
    return rb_str_new2(RVAL2ATTR(self)->name);
}

static VALUE
rb_papi_attribute_get_value_type(VALUE self)
{
    return INT2NUM(RVAL2ATTR(self)->type);
}

static VALUE
c2r_attribute_value_string(papi_attribute_value_t *value)
{
    return rb_str_new2(value->string);
}

static VALUE
c2r_attribute_value_integer(papi_attribute_value_t *value)
{
    return INT2NUM(value->integer);
}

static VALUE
c2r_attribute_value_boolean(papi_attribute_value_t *value)
{
    return value->boolean ? Qtrue : Qfalse;
}

static VALUE
c2r_attribute_value_range(papi_attribute_value_t *value)
{
    return rb_range_new(INT2NUM(value->range.lower),
                        INT2NUM(value->range.upper),
                        FALSE);
}

static VALUE
c2r_attribute_value_resolution(papi_attribute_value_t *value)
{
    return rb_funcall(cResolution,
                      id_new,
                      3,
                      INT2NUM(value->resolution.xres),
                      INT2NUM(value->resolution.yres),
                      INT2NUM(value->resolution.units));
}

static VALUE
c2r_attribute_value_datetime(papi_attribute_value_t *value)
{
    return rb_time_new(value->datetime, 0);
}

static VALUE
c2r_attribute_value_collection(papi_attribute_value_t *value)
{
    VALUE rb_value = rb_ary_new();
    papi_attribute_t **attr;

    for (attr = value->collection; *attr; attr++) {
        rb_ary_push(rb_value, ATTR2RVAL_NO_FREE(*attr));
    }

    return rb_value;
}

static VALUE
c2r_attribute_value_metadata(papi_attribute_value_t *value)
{
    return MD2RVAL(value->metadata);
}

static VALUE
rb_papi_attribute_get_values(VALUE self)
{
    VALUE values = rb_ary_new();
    c2r_attr_value_t c2r_converter = NULL;
    papi_attribute_t *attr;
    papi_attribute_value_t **value;

    attr = RVAL2ATTR(self);
    switch (attr->type) {
      case PAPI_STRING:
        c2r_converter = c2r_attribute_value_string;
        break;
      case PAPI_INTEGER:
        c2r_converter = c2r_attribute_value_integer;
        break;
      case PAPI_BOOLEAN:
        c2r_converter = c2r_attribute_value_boolean;
        break;
      case PAPI_RANGE:
        c2r_converter = c2r_attribute_value_range;
        break;
      case PAPI_RESOLUTION:
        c2r_converter = c2r_attribute_value_resolution;
        break;
      case PAPI_DATETIME:
        c2r_converter = c2r_attribute_value_datetime;
        break;
      case PAPI_COLLECTION:
        c2r_converter = c2r_attribute_value_collection;
        break;
      case PAPI_METADATA:
        c2r_converter = c2r_attribute_value_metadata;
        break;
      default:
        rb_raise(rb_eArgError, "invalid value type %d", attr->type);
        break;
    }

    for (value = attr->values; *value; value++) {
        rb_ary_push(values, c2r_converter(*value));
    }

    return values;
}

papi_attribute_t *
rb_papi_r2c_attribute(VALUE name, VALUE rb_value)
{
    papi_attribute_t *attr;
    papi_attribute_value_t **values;
    papi_attribute_value_t *value;

    attr = ALLOC(papi_attribute_t);
    attr->name = strdup(StringValueCStr(name));

    value = ALLOC(papi_attribute_value_t);

    if (rb_obj_is_kind_of(rb_value, rb_cString)) {
        attr->type = PAPI_STRING;
        value->string = strdup(StringValueCStr(rb_value));
    } else if (rb_obj_is_kind_of(rb_value, rb_cNumeric)) {
        attr->type = PAPI_INTEGER;
        value->integer = NUM2INT(rb_value);
    } else if (rb_value == Qtrue || rb_value == Qfalse) {
        attr->type = PAPI_BOOLEAN;
        value->boolean = RTEST(rb_value);
    } else if (rb_obj_is_kind_of(rb_value, rb_cRange)) {
        attr->type = PAPI_RANGE;
        value->range.lower = NUM2INT(rb_funcall(rb_value, id_begin, 0));
        value->range.upper = NUM2INT(rb_funcall(rb_value, id_end, 0));
    } else if (rb_obj_is_kind_of(rb_value, cResolution)) {
        attr->type = PAPI_RESOLUTION;
        value->resolution.xres = NUM2INT(rb_funcall(rb_value, id_xres, 0));
        value->resolution.yres = NUM2INT(rb_funcall(rb_value, id_yres, 0));
        value->resolution.units = NUM2INT(rb_funcall(rb_value, id_units, 0));
    } else if (rb_obj_is_kind_of(rb_value, rb_cTime)) {
        attr->type = PAPI_DATETIME;
        value->datetime = NUM2LONG(rb_funcall(rb_value, id_tv_sec, 0));
    } else if (rb_obj_is_kind_of(rb_value, rb_cArray)) {
        papi_attribute_t **collection;
        int i, len;

        attr->type = PAPI_COLLECTION;
        len = RARRAY(rb_value)->len;
        collection = ALLOC_N(papi_attribute_t *, len);
        for (i = 0; i < len; i++) {
            VALUE sub_name, sub_value;
            sub_name = rb_ary_entry(RARRAY(rb_value)->ptr[i], 0);
            sub_value = rb_ary_entry(RARRAY(rb_value)->ptr[i], 1);
            collection[i] = rb_papi_r2c_attribute(sub_name, sub_value);
        }
        collection[i] = NULL;
        value->collection = collection;
    } else if (rb_obj_is_kind_of(rb_value, cMetadata)) {
        attr->type = PAPI_METADATA;
        value->metadata = NUM2INT(rb_funcall(rb_value, id_value, 0));
    }

    values = ALLOC_N(papi_attribute_value_t *, 2);
    values[0] = value;
    values[1] = NULL;
    attr->values = values;
    return attr;
}


void
Init_rb_papi_attribute(VALUE mPAPI)
{
    id_begin = rb_intern("begin");
    id_end = rb_intern("end");
    id_xres = rb_intern("xres");
    id_yres = rb_intern("yres");
    id_units = rb_intern("units");
    id_tv_sec = rb_intern("tv_sec");
    id_value = rb_intern("value");

    cAttribute = rb_define_class_under(mPAPI, "Attribute", rb_cData);
    cResolution = rb_const_get(mPAPI, rb_intern("Resolution"));
    cMetadata = rb_const_get(mPAPI, rb_intern("Metadata"));

    rb_define_method(cAttribute, "name", rb_papi_attribute_get_name, 0);
    rb_define_method(cAttribute, "value_type",
                     rb_papi_attribute_get_value_type, 0);
    rb_define_method(cAttribute, "values", rb_papi_attribute_get_values, 0);

    /* PAPI_ATTR_* */
    rb_define_const(cAttribute, "APPEND", INT2NUM(PAPI_ATTR_APPEND));
    rb_define_const(cAttribute, "REPLACE", INT2NUM(PAPI_ATTR_REPLACE));
    rb_define_const(cAttribute, "EXCL", INT2NUM(PAPI_ATTR_EXCL));
}
